@@ -5,7 +5,7 @@
  *
  * Copyright (C) 2005-2006  NTT DATA CORPORATION
  *
- * Version: 1.1.3   2006/07/13
+ * Version: 1.3.1     2006/12/08
  *
  */
 #include <stdio.h>
@@ -19,6 +19,8 @@
 #include <fcntl.h>
 #include <sys/types.h>
 #include <curses.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 
 #define ROOT_NAME "<kernel>"
 
@@ -42,13 +44,13 @@ static int IsCorrectPath(const char *filename, const int may_contain_pattern) {
 				case '$':   /* "\$" */
 				case '+':   /* "\+" */
 				case '?':   /* "\?" */
-					if (may_contain_pattern) continue;
-					break;
 				case '*':   /* "\*" */
-					if (may_contain_pattern) {
-						while (*filename && *filename != '/') filename++;
-						if (!*filename || *filename == '/') continue;
-					}
+				case '@':   /* "\@" */
+				case 'x':   /* "\x" */
+				case 'X':   /* "\X" */
+				case 'a':   /* "\a" */
+				case 'A':   /* "\A" */
+					if (may_contain_pattern) continue;
 					break;
 				case '0':   /* "\ooo" */
 				case '1':
@@ -116,13 +118,23 @@ static inline int strendswith(const char *name, const char *tail) {
 	return len >= 0 && strcmp(name + len, tail) == 0;
 }
 
+static int getch2(void) {
+	static int c0 = 0, c1 = 0, c2 = 0, c3 = 0, len = 0;
+	if (len > 0) { c0 = c1; c1 = c2; c2 = c3; len--; return c0; }
+	c0 = getch(); if (c0 != 0x1B) return c0;
+	c1 = getch(); if (c1 != '[') { len = 1; return c0; }
+	c2 = getch(); if (c2 != '5' && c2 != '6') { len = 2; return c0; }
+	c3 = getch(); if (c3 != '~') { len = 3; return c0; }
+	return c2 == '5' ? KEY_PPAGE : KEY_NPAGE;
+}
+
 static void ReadGenericPolicy(void);
 static void ReadDomainPolicy(void);
 static int FindOrAssignNewDomain(const char *domainname);
 static int FindDomain(const char *domainname);
 static void NormalizeLine(unsigned char *buffer);
 static char *DomainName(const int index);
-static int IsTrustedDomain(const int index);
+static int IsDomainKeeper(const int index);
 static int IsInitializer(const char *filename);
 static int IsInitializerDomain(const int index);
 static int IsVirtualDomain(const int index);
@@ -150,24 +162,25 @@ static int generic_acl_list_count = 0;
 // List for domain policy.
 typedef struct domain_info {
 	char *domainname;
-	int is_virtual;
+	const char **string_ptr; /***** Used by offline mode. *****/
+	int string_count; /***** Used by offline mode. *****/
+	unsigned char is_virtual;
+	unsigned char profile; /***** Used by offline mode. *****/
 } DOMAIN_INFO;
 static DOMAIN_INFO *domain_list = NULL;
 static int domain_list_count = 0;
-static char **trusted_domainnames = NULL;
-static int trusted_domainnames_count = 0;
+static char **domain_keepers = NULL;
+static int domain_keepers_count = 0;
 static char **initializers = NULL;
 static int initializers_count = 0;
 
 ///////////////////////////  ACL HANDLER  //////////////////////////////
 
-static int IsTrustedDomain(const int index) {
+static int IsDomainKeeper(const int index) {
 	int i;
 	char *domainname = domain_list[index].domainname;
-	for (i = 0; i < trusted_domainnames_count; i++) {
-		const int len = strlen(trusted_domainnames[i]);
-		if (strncmp(domainname, trusted_domainnames[i], len)) continue;
-		if (domainname[len] == ' ' || domainname[len] == '\0') return 1;
+	for (i = 0; i < domain_keepers_count; i++) {
+		if (strcmp(domainname, domain_keepers[i]) == 0) return 1;
 	}
 	return 0;
 }
@@ -256,11 +269,75 @@ static void put(void) {
 #define MAXBUFSIZE  8192
 static char buffer[MAXBUFSIZE];
 
+static int offline_mode = 0;
+static int persistent_fd = EOF;
+
+static void SendFD(char *buffer, int *fd) {
+	struct msghdr msg;
+	struct iovec iov = { buffer, strlen(buffer) };
+	char cmsg_buf[CMSG_SPACE(sizeof(int))];
+	struct cmsghdr *cmsg = (struct cmsghdr *) cmsg_buf; 
+	memset(&msg, 0, sizeof(msg));
+	msg.msg_iov = &iov;
+	msg.msg_iovlen = 1;
+	msg.msg_control = cmsg_buf;
+	msg.msg_controllen = sizeof(cmsg_buf);
+	cmsg->cmsg_level = SOL_SOCKET;
+	cmsg->cmsg_type = SCM_RIGHTS;
+	msg.msg_controllen = cmsg->cmsg_len = CMSG_LEN(sizeof(int));
+	memmove(CMSG_DATA(cmsg), fd, sizeof(int));
+	sendmsg(persistent_fd, &msg, 0);
+	close(*fd);
+}
+
+static FILE *open_read(const char *filename) {
+	if (offline_mode) {
+		char buffer[1024];
+		int fd[2];
+		FILE *fp;
+		if (socketpair(PF_UNIX, SOCK_STREAM, 0, fd)) {
+			fprintf(stderr, "socketpair()\n");
+			exit(1);
+		}
+		if (shutdown(fd[0], SHUT_WR) || (fp = fdopen(fd[0], "r")) == NULL) {
+			close(fd[1]); close(fd[0]);
+			exit(1);
+		}
+		memset(buffer, 0, sizeof(buffer));
+		snprintf(buffer, sizeof(buffer) - 1, "GET %s", filename);
+		SendFD(buffer, &fd[1]);
+		return fp;
+	} else {
+		return fopen(filename, "r");
+	}
+}
+
+static int open_write(const char *filename) {
+	if (offline_mode) {
+		char buffer[1024];
+		int fd[2];
+		if (socketpair(PF_UNIX, SOCK_STREAM, 0, fd)) {
+			fprintf(stderr, "socketpair()\n");
+			exit(1);
+		}
+		if (shutdown(fd[0], SHUT_RD)) {
+			close(fd[1]); close(fd[0]);
+			exit(1);
+		}
+		memset(buffer, 0, sizeof(buffer));
+		snprintf(buffer, sizeof(buffer) - 1, "POST %s", filename);
+		SendFD(buffer, &fd[1]);
+		return fd[0];
+	} else {
+		return open(filename, O_WRONLY);
+	}
+}
+
 static void ReadGenericPolicy(void) {
 	FILE *fp;
 	int i, j;
 	while (generic_acl_list_count) free(generic_acl_list[--generic_acl_list_count]);
-	if ((fp = fopen(policy_file, "r")) != NULL) {
+	if ((fp = open_read(policy_file)) != NULL) {
 		if (current_screen == SCREEN_ACL_LIST) {
 			int flag = 0;
 			get();
@@ -314,11 +391,11 @@ static void ReadDomainPolicy(void) {
 	int i, j;
 	while (domain_list_count) free(domain_list[--domain_list_count].domainname);
 	while (initializers_count) free(initializers[--initializers_count]);
-	while (trusted_domainnames_count) free(trusted_domainnames[--trusted_domainnames_count]);
+	while (domain_keepers_count) free(domain_keepers[--domain_keepers_count]);
 	FindOrAssignNewDomain(ROOT_NAME);
 
-	// Load initializer list, trusted domain list.
-	if ((fp = fopen(EXCEPTION_POLICY_FILE, "r")) != NULL) {
+	// Load initializer list, domain_keeper list.
+	if ((fp = open_read(EXCEPTION_POLICY_FILE)) != NULL) {
 		get();
 		while (memset(buffer, 0, sizeof(buffer)), fgets(buffer, sizeof(buffer) - 1, fp) != NULL) {
 			char *cp = strchr(buffer, '\n');
@@ -330,11 +407,11 @@ static void ReadDomainPolicy(void) {
 				if (!IsCorrectPath(buffer, 0)) continue;
 				if ((initializers = (char **) realloc(initializers, (initializers_count + 1) * sizeof(char *))) == NULL
 					|| (initializers[initializers_count++] = strdup(buffer)) == NULL) OutOfMemory();
-			} else if (strncmp(buffer, "trust_domain ", 13) == 0) {
-				memmove(buffer, buffer + 13, strlen(buffer + 13) + 1);
+			} else if (strncmp(buffer, "keep_domain ", 12) == 0) {
+				memmove(buffer, buffer + 12, strlen(buffer + 12) + 1);
 				if (!IsCorrectDomain(buffer)) continue;
-				if ((trusted_domainnames = (char **) realloc(trusted_domainnames, (trusted_domainnames_count + 1) * sizeof(char *))) == NULL
-					|| (trusted_domainnames[trusted_domainnames_count++] = strdup(buffer)) == NULL) OutOfMemory();
+				if ((domain_keepers = (char **) realloc(domain_keepers, (domain_keepers_count + 1) * sizeof(char *))) == NULL
+					|| (domain_keepers[domain_keepers_count++] = strdup(buffer)) == NULL) OutOfMemory();
 			}
 		}
 		put();
@@ -342,7 +419,7 @@ static void ReadDomainPolicy(void) {
 	}
 
 	// Load all domain list.
-	if ((fp = fopen(DOMAIN_POLICY_FILE, "r")) != NULL) {
+	if ((fp = open_read(DOMAIN_POLICY_FILE)) != NULL) {
 		int index = EOF;
 		get();
 		while (memset(buffer, 0, sizeof(buffer)), fgets(buffer, sizeof(buffer) - 1, fp) != NULL) {
@@ -433,7 +510,7 @@ static void ShowList(void) {
 		if (index >= list_item_count[current_screen]) break;
 		if (current_screen == SCREEN_DOMAIN_LIST) {
 			char *sp, *cp;
-			mvprintw(header_lines + i, 0, "%5d: %c%c ", index, IsTrustedDomain(index) ? '#' : ' ', IsInitializerDomain(index) ? '*' : ' ');
+			mvprintw(header_lines + i, 0, "%5d: %c%c ", index, IsDomainKeeper(index) ? '#' : ' ', IsInitializerDomain(index) ? '*' : ' ');
 			sp = DomainName(index);
 			while ((cp = strchr(sp, ' ')) != NULL) { printw("    "); sp = cp + 1; }
 			printw("%s", sp);
@@ -447,6 +524,7 @@ static void ShowList(void) {
 		} else {
 			mvprintw(header_lines + i, 0, "%5d: %s", index, generic_acl_list[index]);
 		}
+		clrtoeol();
 	}
 	ShowCurrent();
 }
@@ -566,7 +644,7 @@ static int GenericListLoop(void) {
 	AdjustCursorPos(generic_acl_list_count);
 	ShowList();
 	while (1) {
-		const int c = getch();
+		const int c = getch2();
 		SaveCursorPos(current_screen);
 		if (c == 'q' || c == 'Q') return MAXSCREEN;
 		if (current_screen == SCREEN_ACL_LIST) {
@@ -590,7 +668,6 @@ static int GenericListLoop(void) {
 			DownArrowKey();
 			break;
 		case KEY_PPAGE:
-		case KEY_DC:
 			PageUpKey();
 			break;
 		case KEY_NPAGE:
@@ -601,7 +678,7 @@ static int GenericListLoop(void) {
 			{
 				const int index = GetCurrent();
 				if (index >= 0 && index < generic_acl_list_count) {
-					const int fd = open(policy_file, O_WRONLY);
+					const int fd = open_write(policy_file);
 					if (fd != EOF) {
 						if (current_screen == SCREEN_ACL_LIST) {
 							write(fd, "select ", 7); write(fd, current_domain, strlen(current_domain)); write(fd, "\n", 1);
@@ -619,10 +696,10 @@ static int GenericListLoop(void) {
 				get();
 				mvprintw(window_height - 1, 0, "Enter new entry> "); clrtoeol(); refresh(); echo();
 				memset(buffer, 0, sizeof(buffer));
-				getnstr(buffer, sizeof(buffer) - 1); noecho();
+				scrollok(stdscr, 1); getnstr(buffer, sizeof(buffer) - 1); scrollok(stdscr, 0); noecho();
 				NormalizeLine(buffer);
 				if (buffer[0]) {
-					const int fd = open(policy_file, O_WRONLY);
+					const int fd = open_write(policy_file);
 					if (fd != EOF) {
 						if (current_screen == SCREEN_ACL_LIST) {
 							write(fd, "select ", 7); write(fd, current_domain, strlen(current_domain)); write(fd, "\n", 1);
@@ -648,7 +725,7 @@ static int DomainListLoop(void) {
 	AdjustCursorPos(domain_list_count);
 	ShowList();
 	while (1) {
-		const int c = getch();
+		const int c = getch2();
 		SaveCursorPos(current_screen);
 		if (c == 'q' || c == 'Q') return MAXSCREEN;
 		if (c == '\t') return SCREEN_SYSTEM_LIST;
@@ -665,7 +742,6 @@ static int DomainListLoop(void) {
 			DownArrowKey();
 			break;
 		case KEY_PPAGE:
-		case KEY_DC:
 			PageUpKey();
 			break;
 		case KEY_NPAGE:
@@ -677,7 +753,7 @@ static int DomainListLoop(void) {
 			input_path:
 				mvprintw(window_height - 1, 0, "Search> "); clrtoeol(); refresh(); echo();
 				memset(history.search_buffer, 0, sizeof(history.search_buffer));
-				getnstr(history.search_buffer, sizeof(history.search_buffer) - 1); noecho();
+				scrollok(stdscr, 1); getnstr(history.search_buffer, sizeof(history.search_buffer) - 1); scrollok(stdscr, 0); noecho();
 				NormalizeLine(history.search_buffer);
 				if (history.search_buffer[0]) {
 					for (i = 0; i < list_item_count[current_screen]; i++) {
@@ -752,10 +828,10 @@ static int DomainListLoop(void) {
 					get();
 					mvprintw(window_height - 1, 0, "Enter new entry> "); clrtoeol(); refresh(); echo();
 					memset(buffer, 0, sizeof(buffer));
-					getnstr(buffer, sizeof(buffer) - 1); noecho();
+					scrollok(stdscr, 1); getnstr(buffer, sizeof(buffer) - 1); scrollok(stdscr, 0); noecho();
 					NormalizeLine(buffer);
 					if (IsCorrectPath(buffer, 0) && !IsInitializer(buffer)) {
-						const int fd = open(DOMAIN_POLICY_FILE, O_WRONLY);
+						const int fd = open_write(DOMAIN_POLICY_FILE);
 						if (fd != EOF) {
 							char *domainname = domain_list[current].domainname;
 							write(fd, domainname, strlen(domainname)); write(fd, " ", 1); write(fd, buffer, strlen(buffer)); write(fd, "\n", 1);
@@ -778,11 +854,11 @@ static int DomainListLoop(void) {
 					clrtoeol();
 					refresh();
 					do {
-						c = getch();
+						c = getch2();
 					} while (!(c == 'Y' || c == 'y' || c == 'N' || c == 'n' || c == EOF));
 					ResizeWindow();
 					if (c == 'Y' || c == 'y') {
-						const int fd = open(DOMAIN_POLICY_FILE, O_WRONLY);
+						const int fd = open_write(DOMAIN_POLICY_FILE);
 						if (fd != EOF) {
 							const int key_len = strlen(key);
 							int index;
@@ -830,6 +906,8 @@ static int DomainListLoop(void) {
 	}
 }
 
+static void policy_daemon(void);
+
 int main(int argc, char *argv[]) {
 	memset(&history, 0, sizeof(history));
 	memset(current_y, 0, sizeof(current_y));
@@ -844,19 +922,66 @@ int main(int argc, char *argv[]) {
 			return 1;
 		}
 	}
-	if (chdir("/proc/ccs/policy/")) {
-		fprintf(stderr, "You can't use this editor for this kernel.\n");
-		return 1;
-	}
 	{
-		const int fd1 = open(SYSTEM_POLICY_FILE, O_RDWR), fd2 = open(EXCEPTION_POLICY_FILE, O_RDWR), fd3 = open(DOMAIN_POLICY_FILE, O_RDWR);
-		if ((fd1 != EOF && write(fd1, "", 0) != 0) || (fd2 != EOF && write(fd2, "", 0) != 0) || (fd3 != EOF && write(fd3, "", 0) != 0)) {
-			fprintf(stderr, "You need to register this program to /root/security/manager.txt and reboot to run this program.\n");
+		char *cp = strrchr(argv[0], '/');
+		if (!cp) cp = argv[0];
+		else cp++;
+		if (strcmp(cp, "editpolicy_offline") == 0) offline_mode = 1;
+	}
+	if (offline_mode) {
+		int fd[2];
+		if (chdir("/etc/ccs/")) {
+			fprintf(stderr, "/etc/ccs/ doesn't exist.\n");
 			return 1;
 		}
-		close(fd1); close(fd2); close(fd3);
+		if (socketpair(PF_UNIX, SOCK_STREAM, 0, fd)) {
+			fprintf(stderr, "socketpair()\n");
+			exit(1);
+		}
+		switch (fork()) {
+		case 0:
+			close(fd[0]);
+			persistent_fd = fd[1];
+			policy_daemon();
+			_exit(0);
+		case -1:
+			fprintf(stderr, "fork()\n");
+			exit(1);
+		}
+		close(fd[1]);
+		persistent_fd = fd[0];
+		{
+			int fd0, fd1, len;
+			if ((fd0 = open("system_policy.txt", O_RDONLY)) != EOF) {
+				fd1 = open_write(SYSTEM_POLICY_FILE);
+				while ((len = read(fd0, buffer, sizeof(buffer))) > 0) write(fd1, buffer, len);
+				close(fd1); close(fd0);
+			}
+			if ((fd0 = open("exception_policy.txt", O_RDONLY)) != EOF) {
+				fd1 = open_write(EXCEPTION_POLICY_FILE);
+				while ((len = read(fd0, buffer, sizeof(buffer))) > 0) write(fd1, buffer, len);
+				close(fd1); close(fd0);
+			}
+			if ((fd0 = open("domain_policy.txt", O_RDONLY)) != EOF) {
+				fd1 = open_write(DOMAIN_POLICY_FILE);
+				while ((len = read(fd0, buffer, sizeof(buffer))) > 0) write(fd1, buffer, len);
+				close(fd1); close(fd0);
+			}
+		}
+	} else {
+		if (chdir("/proc/ccs/policy/")) {
+			fprintf(stderr, "You can't use this editor for this kernel.\n");
+			return 1;
+		}
+		{
+			const int fd1 = open(SYSTEM_POLICY_FILE, O_RDWR), fd2 = open(EXCEPTION_POLICY_FILE, O_RDWR), fd3 = open(DOMAIN_POLICY_FILE, O_RDWR);
+			if ((fd1 != EOF && write(fd1, "", 0) != 0) || (fd2 != EOF && write(fd2, "", 0) != 0) || (fd3 != EOF && write(fd3, "", 0) != 0)) {
+				fprintf(stderr, "You need to register this program to /proc/ccs/policy/manager to run this program.\n");
+				return 1;
+			}
+			close(fd1); close(fd2); close(fd3);
+		}
 	}
-	setenv("TERM", "linux", 1);
 	initscr();
 	cbreak();
 	noecho();
@@ -865,9 +990,11 @@ int main(int argc, char *argv[]) {
 	keypad(stdscr, TRUE);
 	getmaxyx(stdscr, window_height, window_width);
 	while (current_screen < MAXSCREEN) {
-		if (current_screen == SCREEN_DOMAIN_LIST && access(DOMAIN_POLICY_FILE, F_OK)) current_screen = SCREEN_SYSTEM_LIST;
-		else if (current_screen == SCREEN_SYSTEM_LIST && access(SYSTEM_POLICY_FILE, F_OK)) current_screen = SCREEN_EXCEPTION_LIST;
-		else if (current_screen == SCREEN_EXCEPTION_LIST && access(EXCEPTION_POLICY_FILE, F_OK)) current_screen = SCREEN_DOMAIN_LIST;
+		if (!offline_mode) {
+			if (current_screen == SCREEN_DOMAIN_LIST && access(DOMAIN_POLICY_FILE, F_OK)) current_screen = SCREEN_SYSTEM_LIST;
+			else if (current_screen == SCREEN_SYSTEM_LIST && access(SYSTEM_POLICY_FILE, F_OK)) current_screen = SCREEN_EXCEPTION_LIST;
+			else if (current_screen == SCREEN_EXCEPTION_LIST && access(EXCEPTION_POLICY_FILE, F_OK)) current_screen = SCREEN_DOMAIN_LIST;
+		}
 		ResizeWindow();
 		if (current_screen == SCREEN_DOMAIN_LIST) current_screen = DomainListLoop();
 		else current_screen = GenericListLoop();
@@ -876,5 +1003,343 @@ int main(int argc, char *argv[]) {
 	move(0, 0);
 	refresh();
 	endwin();
+	if (offline_mode) {
+		int fd, len;
+		FILE *fp;
+		time_t now = time(NULL);
+		struct tm *tm = localtime(&now);
+		char filename[1024];
+		memset(filename, 0, sizeof(filename));
+		snprintf(filename, sizeof(filename) - 1, "system_policy.%02d-%02d-%02d.%02d:%02d:%02d.txt", tm->tm_year % 100, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
+		if ((fd = open(filename, O_WRONLY | O_CREAT, 0600)) != EOF) {
+			if ((fp = open_read(SYSTEM_POLICY_FILE)) != NULL) {
+				get(); while ((len = fread(buffer, 1, sizeof(buffer), fp)) > 0) write(fd, buffer, len); put();
+				close(fd); fclose(fp);
+				unlink("system_policy.txt");
+				symlink(filename, "system_policy.txt");
+			}
+		}
+		snprintf(filename, sizeof(filename) - 1, "exception_policy.%02d-%02d-%02d.%02d:%02d:%02d.txt", tm->tm_year % 100, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
+		if ((fd = open(filename, O_WRONLY | O_CREAT, 0600)) != EOF) {
+			if ((fp = open_read(EXCEPTION_POLICY_FILE)) != NULL) {
+				get(); while ((len = fread(buffer, 1, sizeof(buffer), fp)) > 0) write(fd, buffer, len); put();
+				close(fd); fclose(fp);
+				unlink("exception_policy.txt");
+				symlink(filename, "exception_policy.txt");
+			}
+		}
+		snprintf(filename, sizeof(filename) - 1, "domain_policy.%02d-%02d-%02d.%02d:%02d:%02d.txt", tm->tm_year % 100, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
+		if ((fd = open(filename, O_WRONLY | O_CREAT, 0600)) != EOF) {
+			if ((fp = open_read(DOMAIN_POLICY_FILE)) != NULL) {
+				get(); while ((len = fread(buffer, 1, sizeof(buffer), fp)) > 0) write(fd, buffer, len); put();
+				close(fd); fclose(fp);
+				unlink("domain_policy.txt");
+				symlink(filename, "domain_policy.txt");
+			}
+		}
+	}
 	return 0;
 }
+
+
+
+/***** The code below is used only when offline mode. *****/
+
+/* Copied from kernel source. */
+static inline unsigned long partial_name_hash(unsigned long c, unsigned long prevhash) {
+	return (prevhash + (c << 4) + (c >> 4)) * 11;
+}
+
+/* Copied from kernel source. */
+static inline unsigned int full_name_hash(const unsigned char *name, unsigned int len) {
+	unsigned long hash = 0;
+	while (len--) hash = partial_name_hash(*name++, hash);
+	return (unsigned int) hash;
+}
+
+#define PAGE_SIZE  4096
+
+static char *alloc_element(const unsigned int size) {
+	static char *buf = NULL;
+	static unsigned int buf_used_len = PAGE_SIZE;
+	char *ptr = NULL;
+	if (size > PAGE_SIZE) return NULL;
+	if (buf_used_len + size > PAGE_SIZE) {
+		if ((ptr = malloc(PAGE_SIZE)) == NULL) OutOfMemory();
+		buf = ptr;
+		memset(buf, 0, PAGE_SIZE);
+		buf_used_len = size;
+		ptr = buf;
+	} else if (size) {
+		int i;
+		ptr = buf + buf_used_len;
+		buf_used_len += size;
+		for (i = 0; i < size; i++) if (ptr[i]) OutOfMemory();
+	}
+	return ptr;
+}
+
+#define MAX_HASH 256
+
+typedef struct name_entry {
+	struct name_entry *next; /* Pointer to next record. NULL if none.             */
+	unsigned int hash;       /* hash and length                                   */
+	const char *name;        /* Text form of filename and domainname. Never NULL. */
+} NAME_ENTRY;
+
+typedef struct free_memory_block_list {
+	struct free_memory_block_list *next; /* Pointer to next record. NULL if none. */
+	char *ptr;                           /* Pointer to a free area.               */
+	int len;                             /* Length of the area.                   */
+} FREE_MEMORY_BLOCK_LIST;
+
+static const char *SaveName(const char *name) {
+	static FREE_MEMORY_BLOCK_LIST fmb_list = { NULL, NULL, 0 };
+	static NAME_ENTRY name_list[MAX_HASH]; /* The list of names. */
+	NAME_ENTRY *ptr, *prev = NULL;
+	unsigned int hash;
+	FREE_MEMORY_BLOCK_LIST *fmb = &fmb_list;
+	int len;
+	static int first_call = 1;
+	if (!name) return NULL;
+	len = strlen(name) + 1;
+	if (len > PAGE_SIZE) {
+		fprintf(stderr, "ERROR: Name too long for SaveName().\n");
+		return NULL;
+	}
+	hash = full_name_hash((const unsigned char *) name, len - 1);
+	if (first_call) {
+		int i;
+		first_call = 0;
+		memset(&name_list, 0, sizeof(name_list));
+		for (i = 0; i < MAX_HASH; i++) name_list[i].name = "/";
+	}
+	ptr = &name_list[hash % MAX_HASH];
+	hash ^= len; /* The hash % MAX_HASH are always same for ptr->hash, so embed length into the hash value. */
+	while (ptr) {
+		if (hash == ptr->hash && strcmp(name, ptr->name) == 0) goto out;
+		prev = ptr; ptr = ptr->next;
+	}
+	while (len > fmb->len) {
+		if (fmb->next) {
+			fmb = fmb->next;
+		} else {
+			char *cp;
+			if ((cp = (char *) malloc(PAGE_SIZE)) == NULL || (fmb->next = (FREE_MEMORY_BLOCK_LIST *) alloc_element(sizeof(FREE_MEMORY_BLOCK_LIST))) == NULL) OutOfMemory();
+			memset(cp, 0, PAGE_SIZE);
+			fmb = fmb->next;
+			fmb->ptr = cp;
+			fmb->len = PAGE_SIZE;
+		}
+	}
+	if ((ptr = (NAME_ENTRY *) alloc_element(sizeof(NAME_ENTRY))) == NULL) OutOfMemory();
+	memset(ptr, 0, sizeof(NAME_ENTRY));
+	ptr->hash = hash;
+	ptr->name = fmb->ptr;
+	memmove(fmb->ptr, name, len);
+	fmb->ptr += len;
+	fmb->len -= len;
+	prev->next = ptr; /* prev != NULL because name_list is not empty. */
+	if (fmb->len == 0) {
+		FREE_MEMORY_BLOCK_LIST *ptr = &fmb_list;
+		while (ptr->next != fmb) ptr = ptr->next; ptr->next = fmb->next;
+	}
+ out:
+	return ptr ? (const char *) ptr->name : NULL;
+}
+
+static void DeleteDomain(const int index) {
+	if (index > 0 && index < domain_list_count) {
+		int i;
+		free(domain_list[index].string_ptr);
+		for (i = index; i < domain_list_count - 1; i++) domain_list[i] = domain_list[i + 1];
+		domain_list_count--;
+	}
+}
+
+static int AddStringEntry(const char *entry, const int index) {
+	const char **acl_ptr;
+	int acl_count;
+	const char *cp;
+	int i;
+	if (index < 0 || index >= domain_list_count) {
+		fprintf(stderr, "AddStringEntry: ERROR: domain is out of range.\n");
+		return -EINVAL;
+	}
+	if (!entry || !*entry) return -EINVAL;
+	if ((cp = SaveName(entry)) == NULL) OutOfMemory();
+
+	acl_ptr = domain_list[index].string_ptr;
+	acl_count = domain_list[index].string_count;
+
+	// Check for the same entry.
+	for (i = 0; i < acl_count; i++) {
+		// Faster comparison, for they are SaveName'd.
+		if (cp == acl_ptr[i]) return 0;
+	}
+
+	if ((acl_ptr = (const char **) realloc(acl_ptr, (acl_count + 1) * sizeof(const char *))) == NULL) OutOfMemory();
+	acl_ptr[acl_count++] = cp;
+	domain_list[index].string_ptr = acl_ptr;
+	domain_list[index].string_count = acl_count;
+	return 0;
+}
+
+static int DelStringEntry(const char *entry, const int index) {
+	const char **acl_ptr;
+	int acl_count;
+	const char *cp;
+	int i;
+	if (index < 0 || index >= domain_list_count) {
+		fprintf(stderr, "DelStringEntry: ERROR: domain is out of range.\n");
+		return -EINVAL;
+	}
+	if (!entry || !*entry) return -EINVAL;
+	if ((cp = SaveName(entry)) == NULL) OutOfMemory();
+
+	acl_ptr = domain_list[index].string_ptr;
+	acl_count = domain_list[index].string_count;
+
+	for (i = 0; i < acl_count; i++) {
+		// Faster comparison, for they are SaveName'd.
+		if (cp != acl_ptr[i]) continue;
+		domain_list[index].string_count--;
+		for (; i < acl_count - 1; i++) acl_ptr[i] = acl_ptr[i + 1];
+		return 0;
+	}
+	return -ENOENT;
+}
+
+static void policy_daemon(void) {
+	while (1) {
+		static const char **exception_list = NULL, **system_list = NULL;
+		static int exception_list_count = 0, system_list_count = 0;
+		FILE *fp;
+		{
+			struct msghdr msg;
+			struct iovec iov = { buffer, sizeof(buffer) - 1 };
+			char cmsg_buf[CMSG_SPACE(sizeof(int))];
+			struct cmsghdr *cmsg = (struct cmsghdr *) cmsg_buf;
+			memset(&msg, 0, sizeof(msg));
+			msg.msg_iov = &iov;
+			msg.msg_iovlen = 1;
+			msg.msg_control = cmsg_buf;
+			msg.msg_controllen = sizeof(cmsg_buf);
+			memset(buffer, 0, sizeof(buffer));
+			errno = 0;
+			if (recvmsg(persistent_fd, &msg, 0) > 0 &&
+				(cmsg = CMSG_FIRSTHDR(&msg)) != NULL &&
+				cmsg->cmsg_level == SOL_SOCKET &&
+				cmsg->cmsg_type == SCM_RIGHTS &&
+				cmsg->cmsg_len == CMSG_LEN(sizeof(int))) {
+				const int fd = * (int *) CMSG_DATA(cmsg);
+				if ((fp = fdopen(fd, "w+")) == NULL) {
+					close(fd);
+					continue;
+				}
+			} else {
+				break;
+			}
+		}
+		if (strncmp(buffer, "POST ", 5) == 0) {
+			if (strcmp(buffer + 5, "domain_policy") == 0) {
+				int index = EOF;
+				while (memset(buffer, 0, sizeof(buffer)), fgets(buffer, sizeof(buffer) - 1, fp) != NULL) {
+					int is_delete = 0, is_select = 0;
+					char *cp = strchr(buffer, '\n');
+					if (cp) *cp = '\0';
+					else if (!feof(fp)) break;
+					NormalizeLine(buffer);
+					if (strncmp(buffer, "delete ", 7) == 0) {
+						is_delete = 1;
+						memmove(buffer, buffer + 7, strlen(buffer + 7) + 1);
+					} else if (strncmp(buffer, "select ", 7) == 0) {
+						is_select = 1;
+						memmove(buffer, buffer + 7, strlen(buffer + 7) + 1);
+					}
+					if (IsDomainDef(buffer)) {
+						if (is_delete) {
+							index = FindDomain(buffer);
+							if (index > 0) DeleteDomain(index);
+							index = EOF;
+						} else if (is_select) {
+							index = FindDomain(buffer);
+						} else {
+							index = FindOrAssignNewDomain(buffer);
+						}
+					} else if (index >= 0 && buffer[0]) {
+						unsigned int profile;
+						if (sscanf(buffer, "use_profile %u", &profile) == 1) {
+							domain_list[index].profile = (unsigned char) profile;
+						} else if (is_delete) {
+							DelStringEntry(buffer, index);
+						} else {
+							AddStringEntry(buffer, index);
+						}
+					}
+				}
+			} else if (strcmp(buffer + 5, "exception_policy") == 0) {
+				while (memset(buffer, 0, sizeof(buffer)), fgets(buffer, sizeof(buffer) - 1, fp) != NULL) {
+					char *cp = strchr(buffer, '\n');
+					if (cp) *cp = '\0';
+					else if (!feof(fp)) break;
+					NormalizeLine(buffer);
+					if (!buffer[0]) continue;
+					if (strncmp(buffer, "delete ", 7) == 0) {
+						int i;
+						memmove(buffer, buffer + 7, strlen(buffer + 7) + 1);
+						for (i = 0; i < exception_list_count; i++) {
+							if (strcmp(exception_list[i], buffer)) continue;
+							for (exception_list_count--; i < exception_list_count; i++) exception_list[i] = exception_list[i + 1];
+							break;
+						}
+					} else {
+						if ((exception_list = (const char **) realloc(exception_list, (exception_list_count + 1) * sizeof(char *))) == NULL
+							|| (exception_list[exception_list_count++] = SaveName(buffer)) == NULL) OutOfMemory();
+					}
+				}
+			} else if (strcmp(buffer + 5, "system_policy") == 0) {
+				while (memset(buffer, 0, sizeof(buffer)), fgets(buffer, sizeof(buffer) - 1, fp) != NULL) {
+					char *cp = strchr(buffer, '\n');
+					if (cp) *cp = '\0';
+					else if (!feof(fp)) break;
+					NormalizeLine(buffer);
+					if (!buffer[0]) continue;
+					if (strncmp(buffer, "delete ", 7) == 0) {
+						int i;
+						memmove(buffer, buffer + 7, strlen(buffer + 7) + 1);
+						for (i = 0; i < system_list_count; i++) {
+							if (strcmp(system_list[i], buffer)) continue;
+							for (system_list_count--; i < system_list_count; i++) system_list[i] = system_list[i + 1];
+							break;
+						}
+					} else {
+						if ((system_list = (const char **) realloc(system_list, (system_list_count + 1) * sizeof(char *))) == NULL
+							|| (system_list[system_list_count++] = SaveName(buffer)) == NULL) OutOfMemory();
+					}
+				}
+			}
+		} else if (strncmp(buffer, "GET ", 4) == 0) {
+			if (strcmp(buffer + 4, "domain_policy") == 0) {
+				int i, j;
+				for (i = 0; i < domain_list_count; i++) {
+					const char **string_ptr = domain_list[i].string_ptr;
+					const int string_count = domain_list[i].string_count;
+					fprintf(fp, "%s\nuse_profile %u\n\n", domain_list[i].domainname, domain_list[i].profile);
+					for (j = 0; j < string_count; j++) {
+						fprintf(fp, "%s\n", string_ptr[j]);
+					}
+					fprintf(fp, "\n");
+				}
+			} else if (strcmp(buffer + 4, "exception_policy") == 0) {
+				int i;
+				for (i = 0; i < exception_list_count; i++) fprintf(fp, "%s\n", exception_list[i]);
+			} else if (strcmp(buffer + 4, "system_policy") == 0) {
+				int i;
+				for (i = 0; i < system_list_count; i++) fprintf(fp, "%s\n", system_list[i]);
+			}
+		}
+		fclose(fp);
+	}
+	_exit(0);
+}
